using Interop.NeoTicker;
using System.Collections;
using System.Reflection;
using System;

namespace KreslikLib.Neo.JukeBox
{
    // usage example:

    //public class EntryPointClass : IDLIndicator
    //{
    //    NTIndicatorJukeBox<MyIDLIndicator, MyCommonObject> JukeBox;

    //    public EntryPointClass()
    //    {
    //        JukeBox = new NTIndicatorJukeBox<MyIDLIndicator, MyCommonObject>();
    //    }

    //    public double IDLCallEx(NTIndicatorObjects N)
    //    {
    //        return JukeBox.IDLCallSong(N);
    //    }
    //}

    public class NTIndicatorJukeBox<T, U>
        where T : Song<U>
        where U : class
    {
        private ArrayList Songs;
        private U CommonObj;

        private int UniqueID;
        private int SongIndex;

        public NTIndicatorJukeBox()
        {
            Songs = new ArrayList();

            Type commonObjectType = typeof(U);
            CommonObj = (U)commonObjectType.InvokeMember(commonObjectType.Name,
                    BindingFlags.CreateInstance | BindingFlags.Instance | BindingFlags.Public,
                    null, null, null);
        }

        public double IDLCallSong(NTIndicatorObjects N)
        {
            UniqueID = N.ItSelf().UniqueId;
            SongIndex = GetSongIndexByID(UniqueID);

            if (SongIndex < 0)
            {
                Type songType = typeof(T);
                T newSong = (T)songType.InvokeMember(songType.Name,
                    BindingFlags.CreateInstance | BindingFlags.Instance | BindingFlags.Public,
                    null, null, null);

                SongIndex = Songs.Add(newSong);

                ((Song<U>)Songs[SongIndex]).SetupSong(UniqueID, CommonObj, Songs);
            }

            return ((Song<U>)Songs[SongIndex]).IDLCall(N);
        }

        private int GetSongIndexByID(int ID)
        {
            for (int i = 0; i < Songs.Count; i++)
            {
                if (Songs[i] != null && ((Song<U>)Songs[i]).UniqueID == ID)
                {
                    return i;
                }
            }

            return -1;
        }
    }

    /// <summary>
    /// Song is a class that an IDL indicator must inherit. It represents a unique instance of the IDL indicator.
    /// </summary>
    /// <typeparam name="V">V is a type of class that is shared among all unique indicator instances.</typeparam>
    public abstract class Song<V>
        where V : class
    {
        private int uniqueID;
        private V commonObj;
        private ArrayList songs;
        private bool locked;

        public Song()
        {
            // unlock the class so that we can set it up:
            locked = false;
        }

        public void SetupSong(int UniqueID, V CommonObj, ArrayList Songs)
        {
            // make sure that we don't accidentally assign a value or reference
            // to these vital objects later:
            if (!locked)
            {
                uniqueID = UniqueID;
                commonObj = CommonObj;
                songs = Songs;
                locked = true;
            }
            else
            {
                throw new Exception("The " + this.GetType().Name + " object is locked.");
            }
        }

        /// <summary>
        /// Read only property UniqueID is set upon the Song's instantiation.
        /// </summary>
        public int UniqueID
        {
            get { return uniqueID; }
        }

        /// <summary>
        /// Read only property CommonObj is set upon the Song's instantiation.
        /// </summary>
        protected V CommonObj
        {
            get { return commonObj; }
        }

        /// <summary>
        /// Read only property Songs is set upon the Song's instantiation.
        /// </summary>
        protected ArrayList Songs
        {
            get { return songs; }
        }

        public abstract double IDLCall(NTIndicatorObjects N);
    }
}